<?php
/*
	mobiki - the wiki for the mobile world
	Author:  Martin Doering <martin@datenbrei.de>
	Release: 2014-12-18
	Project: http://mobiki.datenbrei.de
	License: http://mobiki.datenbrei.de/index.php?page=License
*/

	// Set internal character encoding to 'UTF-8' - needed for some functions below
	mb_internal_encoding("UTF-8");

	// load password and constants - text MUST be in UTF-8!!!
	require('config.php'); 
	
	$req = new Request();
	
	// make a sanitized URL from a string
	function html($string) { return htmlspecialchars($string); }
	
	// make a sanitized URL from a string
	function url($string) { return urlencode($string); }
	
	// get filename from pagename
	function filename($pagename) { return preg_replace('/\+/', ' ', urlencode($pagename)) . '.txt'; }

	// get pagename from filename
	function pagename($filename) { return urldecode(mb_substr($filename, 0, mb_strlen(preg_replace('/ /', '+', $filename)) -4)); }

	class Request {
		public $host;
		public $script;
		public $page; 
		public $cookie; 
		public $query;
		public $action;
		private $message;
		private $password;
		private $content;
		private $domain;
		
		// Here starts the real thing
		function __construct() {
			$this->cookie = (isset($_COOKIE['mobiki'])) ? $_COOKIE['mobiki'] : NULL; 
			$this->host = $_SERVER['HTTP_HOST'];
			$this->domain = ($_SERVER['HTTP_HOST'] != 'localhost') ? $_SERVER['HTTP_HOST'] : FALSE;			
			if ($_SERVER["REQUEST_METHOD"] == "POST") {
				$this->content = (isset($_POST['content'])) ? $_POST['content'] : NULL;
				$this->password = filter_input(INPUT_POST, "password", FILTER_SANITIZE_STRING,FILTER_FLAG_STRIP_LOW);
				$this->page = htmlspecialchars_decode(filter_input(INPUT_POST, "page", FILTER_UNSAFE_RAW, FILTER_FLAG_STRIP_LOW));
				$this->post();
			}
			elseif ($_SERVER["REQUEST_METHOD"] == "GET") {
				$this->script = $_SERVER['PHP_SELF'];
				$this->page = (isset($_GET['page'])) ? $_GET['page'] : NULL; 
				$this->action = filter_input(INPUT_GET, "action", FILTER_SANITIZE_STRING, FILTER_FLAG_STRIP_LOW);
				$this->message = (filter_input(INPUT_GET, "error", FILTER_VALIDATE_INT)) ? PWD_ERROR : "";
				$this->query = filter_input(INPUT_GET, "query", FILTER_SANITIZE_STRING, FILTER_FLAG_STRIP_LOW);
				$this->get();
			}
		}

		// Write changes to page, if possible and authorized
		function post() {

			// Initial PASSWORD test to enshure no prepared POST hack is possible 
			if (PASSWORD == 'unsetpassword' && $this->page == "" && $this->password != "") {
				$config = file_get_contents('config.php');
				if ($config) {
					$config = str_replace('unsetpassword', crypt($this->password), $config);
					file_put_contents('config.php', $config);
				}
				die(header('Location: index.php'));
			}			

			// Login
			if (crypt($this->password, PASSWORD) == PASSWORD) {
				setcookie('mobiki', PASSWORD, time() + 365*24*3600); // one year
				header("Location: index.php?page=" . url($this->page));
			}

			// Write page, if possible and authorized
			elseif ($this->cookie == PASSWORD) {
				$filename = filename($this->page);
				if ($this->content <> "") {
					if (! $file = fopen(DIRECTORY . $filename, "w"))
						die ("Could not write page!");
					fputs($file, trim($this->content));
					fclose($file);
					header("Location: index.php?page=" . url($this->page));
				}
				else {
					// saving an empty page means: delete it
					unlink(DIRECTORY . $filename);
					header('Location: index.php');
				}
			}
			
			// Password given, but wrong
			else { 
				header("Location: index.php?page=" . url($this->page) . "&action=edit&error=1");
			}

		}
		
		function get() {
			$page = new Page();
			
			if (PASSWORD == 'unsetpassword') $page->set_initial_password();
			elseif ($this->action == 'rss') new Feed($this);
			elseif ($this->action == 'search') $page->print_search_result($this);
			elseif ($this->action == 'hashtag') $page->print_hashtags_found($this);
			elseif ($this->action == 'recent') $page->show_recent_changes($this);
			elseif ($this->action == 'login') $page->show_login_page($this);
			else {
				$page->load_content($this);
				if ($this->action == 'edit') 
					$page->prepare_edit_form($this);
				else
					$page->prepare_html_output($this);
			}

			// drop cookie in browser and actual environment
			if ($this->action == 'logout') {
				setcookie('mobiki', '', time() -3600); 
				$this->cookie = NULL;
			}
			$this->print_html($page);
		}

		function print_html($page) {
			// Read template
			if (! $file = @fopen("template.html", "r"))
				die ("'template.html' is missing!");
			$html = fread($file, filesize("template.html"));
			fclose($file);

			// Put values into template
			if (($this->cookie) && is_writable(DIRECTORY . filename($this->page)))
				$html = str_replace('{EDIT}', '<a id="edit" href="index.php?page=' . url($this->page) . '&amp;action=edit">' . html(EDIT_BUTTON) . '</a>', $html);
			else
				$html = str_replace('{EDIT}', '', $html);
			
			$html = str_replace('{PAGE_TITLE}', html($page->title), $html);
			$html = str_replace('{TIME}', '<span id="time">' . html(date(TIME_FORMAT, $page->time)) . '</span>', $html);
			$html = str_replace('{CONTENT}', $page->content, $html);
			
			$html = str_replace('{HOST}', $this->host, $html);
			$html = str_replace('{SCRIPT}', $this->script, $html);
			$html = str_replace('{AUTHOR}', WIKI_AUTHOR, $html);
			$html = str_replace('{DESCRIPTION}', WIKI_DESCRIPTION, $html);
			$html = str_replace('{HOME}', '<a id="home" title="' . WIKI_DESCRIPTION . '" href="index.php">' . html(HOME_BUTTON) . '</a>', $html);
			$html = str_replace('{FEED}', '<a id="feed" href="index.php?action=rss">' . html(FEED_BUTTON) . '</a>', $html);
			$html = str_replace('{WIKI_TITLE}', html(WIKI_TITLE), $html);
			$html = str_replace('{SEARCH}', 	'<form method="get" action="index.php">' . 
												'<input type="hidden" name="action" value="search" />' . 
												'<input id="search" type="text" name="query" value="' . html($this->query) . '" placeholder="' . url(SEARCH_BUTTON) . '" />' . 
												'</form>', $html);
			$html = str_replace('{NEWS}', '<a id="news" href="index.php?action=recent">' . html(RECENT_CHANGES) . '</a>', $html);
			$html = str_replace('{LOGOUT}', ($this->cookie) ? '<a id="logout" href="index.php?page=' . url($this->page) . '&amp;action=logout">' . html(LOGOUT_BUTTON) . '</a>' 
															: '<a id="login" href="index.php?page=' . url($this->page) . '&amp;action=login">' . html(LOGIN_BUTTON) . '</a>', $html);
			$html = str_replace('{INFO}', '<a id="info" href="index.php?page=' . url(INFO_PAGE) . '">' . html(INFO_BUTTON) . '</a>', $html);
			$html = str_replace('{MESSAGE}', '<span id="message">' . html($this->message) . '</span>', $html);
		
			// Print page            
			header('Content-type: text/html; charset=UTF-8');
			echo $html;
		}
	}
	
	class Page {
		public $title = "";
		public $time;
		public $content = "";

		function __construct() {
			$this->time = time();
		}

		// Prepare initial password form
		function set_initial_password() {
			$this->title = PWD_SET;
			$this->content = '<form method="post" action="index.php">';
			$this->content .= '<input type="password" name="password" placeholder="' . PWD_SET . '" />';
			$this->content .= '</form>';
		}
	
		// Search with sanatized query
		function print_search_result($req) {
			$this->title = SEARCH_RESULTS . ' "' . $req->query . '"';
			$query = preg_replace('/(\*|\+|\\\\|\/)/u', ".*", $req->query);
			$dir = opendir(DIRECTORY);
			$RESULT = "";
			while ($file = readdir($dir)) {
				if (preg_match("/.txt/", $file)) {
					$pagename= pagename($file);
					if (preg_match('/' . $query . '/i', $pagename)) {
						$RESULT .= "<a href=\"index.php?page=" . url($pagename) . "\">" . html($pagename) . "</a><br />";
					}
				}
			}
			if ($RESULT <> "")
				$this->content = "<h2>". RESULT_TITLE . "</h2>" . $RESULT;

			$dir = opendir(DIRECTORY);
			$RESULT="";
			while ($file = readdir($dir)) {
				if (preg_match("/.txt/", $file)) {
					if (preg_match('/' . $query . '/i', file_get_contents(DIRECTORY . $file))) {
						$pagename = pagename($file);
						$RESULT .= '<a href="index.php?page=' . url($pagename) . '">' . html($pagename) . '</a><br />';
					}
				}
			}
			if ($RESULT <> "")
				$this->content .= "<br /><h2>" . RESULT_TEXT . "</h2>" . $RESULT;
		}

		// Search with sanatized query
		function print_hashtags_found($req) {
			$req->query = "#" . preg_replace('/(\*|\+|\\\\|\/)/', "", $req->query);
			$this->title = HASHTAG_RESULTS . " \"" . $req->query . "\"";
			$dir = opendir(DIRECTORY);
			while ($file = readdir($dir)) {
				if (preg_match("/.txt/", $file)) {
					if (preg_match('/' . $req->query . '/i', file_get_contents(DIRECTORY . $file))) {
						$pagename = pagename($file);
						$this->content .= "<a href=\"index.php?page=" . url($pagename) . "\">" . html($pagename) . "</a><br />";
					}
				}
			}
		}

		// Recent changes
		function show_recent_changes($req) {
			$this->title = RECENT_CHANGES;
			$dir = opendir(getcwd() . "/" . DIRECTORY);
			while ($file = readdir($dir))
				if (preg_match("/.txt/", $file))
					$filetime[$file] = filemtime(DIRECTORY . $file);
			arsort($filetime);
			$filetime = array_slice($filetime, 0, 20);
			foreach ($filetime as $filename => $timestamp) {
				$pagename = pagename($filename);
				$this->content .= date(TIME_FORMAT, $timestamp) . ' <a href="index.php?page=' . url($pagename) . '">' . html($pagename) . '</a><br />';
			}
		}

		// Prepare page formatting
		function load_content($req) {
			if (!$req->page)
				$req->page = START_PAGE;
			$this->title = $req->page;
			
			// Read page contents and time of last change
			$filename = filename($req->page);
			if (($file = @fopen(DIRECTORY . $filename, "r")) || $req->action <> "") {
				$this->time = filemtime(DIRECTORY . $filename);
				$this->content = "\n" . @fread($file, @filesize(DIRECTORY . $filename)) . "\n";
				@fclose($file);
			}
			else {
				$req->action = 'edit';
			}
		}	

		// Login page
		function show_login_page($req) {
			$this->title = LOGIN_TITLE;
			$this->content = '<form method="post" action="index.php">';
			$this->content .= '<input type="hidden" name="page" value="' . html($req->page) . '" /><br />';
			$this->content .= '<input type="password" name="password" placeholder="' . PWD_MESSAGE . '" />';
			$this->content .= '</form>';
			$this->message = PWD_MESSAGE;
		}
		
		// Editing the page
		function prepare_edit_form($req) {
			$this->content = '<form method="post" action="index.php"><textarea name="content" rows="20">' . html($this->content) . '</textarea>';
			$this->content .= '<input type="hidden" name="page" value="' . html($req->page) . '" /><br />';
			$this->content .= (PASSWORD != "" && $req->cookie == PASSWORD) ? ' <input id="save" type="submit" value="' . SAVE_BUTTON . '" />' : '<br />';
			$this->content .= '</form>';
		}
		
		// Convert wiki text to HTML
		function prepare_html_output($req) {
			// Regex for mail and http URL validation and internal link
			define('RGX_MAIL', '[0-9a-zA-Z\./~\-_]+@[0-9a-z\./~\-_]+');
			define('RGX_URL', 'https?://[0-9a-zA-Z\.\&amp;\#\:/~\-_%=?]*');
			define('RGX_WIKI', '[^\]]+');
			
			$ref = new Reference();
			$this->content = preg_replace_callback('#{(.+)}#Ums', array($ref, 'pre'), $this->content); // save pre text
			$this->content = preg_replace_callback('#\[(' . RGX_WIKI . ')\|(' . RGX_URL . ')\]#Uu', array($ref, 'link'), $this->content);  //external link
			$this->content = preg_replace_callback('#(' . RGX_URL . '\.(jpg|gif|png))#i', array($ref,'img'), $this->content); // external image
			$this->content = preg_replace_callback('#(' . RGX_URL . ')#i', array($ref, 'url'), $this->content); //external URL
			$this->content = preg_replace_callback("/#(\pL+)/u", array($ref, 'hashtag'), $this->content); // hashtag
			$this->content = preg_replace_callback("/\[(" . RGX_WIKI . ")\]/Uu", array($ref, 'pagename'), $this->content); // Internal page
			$this->content = html($this->content);
			$this->content = preg_replace('#(' . RGX_MAIL . ')#i', '<a href="mailto:$0">$0</a>', $this->content); // mail address URL
			$this->content = preg_replace('/^---+(\r\n|\r|\n)/m', '<hr>', $this->content);
			$this->content = preg_replace('/^!(.+)(\n)/Um', '<h2>$1</h2>$2', $this->content);
			$this->content = preg_replace('/^[\*\-] *(.+)\n/Um', "<ul><li>$1</li></ul>", $this->content);
			$this->content = preg_replace('/^\d\. *(.+)\n/Um', "<ol><li>$1</li></ol>", $this->content);
			$this->content = preg_replace('/(<\/ol><ol>|<\/ul><ul>)/', "", $this->content);
			$this->content = preg_replace_callback('/{([a-f0-9]{32})}/', array($ref, 'get'), $this->content); // get back temporarily stored strings
			$this->content = preg_replace('/\n/', "<br />\n", $this->content);
			$this->content = preg_replace('#(</h[123]>)<br />#', "$1", $this->content);
			$this->content = preg_replace("/'''''(.*)'''''/Um", '<b><i>$1</i></b>', $this->content);
			$this->content = preg_replace("/'''(.*)'''/Um", '<b>$1</b>', $this->content);
			$this->content = preg_replace("/''(.*)''/Um", '<i>$1</i>', $this->content); 
			$this->content = substr($this->content, 6, strlen($this->content) - 6);
		}
	}

	class Feed {
		private $xml; 
		
		// Generate RSS 2.0 feed
		function __construct($req) {
			$xml = "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n";
			$xml .= "<rss version=\"2.0\">\n";
			$xml .= "<channel>\n";
			$xml .= "<title>" . html(WIKI_TITLE) . "</title>\n";
			$xml .= "<link>http://" . $req->host . $req->script . "</link>\n";
			$xml .= "<description>" . html(WIKI_DESCRIPTION) . "</description>\n";
			$xml .= "<pubDate>" . html(date(DATE_RSS)) . "</pubDate>\n";
			$xml .= "<generator>mobiki - the wiki for the mobile world</generator>\n";
			$dir = opendir(getcwd() . "/" . DIRECTORY);
			while ($file = readdir($dir))
				if (preg_match("/.txt$/", $file))
					$filetime[$file] = filectime(DIRECTORY . $file);
			arsort($filetime);
			$filetime = array_slice($filetime, 0, 10);
			foreach ($filetime as $filename => $timestamp) {
				$pagename = pagename($filename);
				$xml .= "<item>\n";
				$xml .= "<title>" . html($pagename) . "</title>\n";
				$xml .= "<link>http://" . $req->host . $req->script . "?page=" . url($pagename) . "</link>\n";
				$xml .= "<guid>http://" . $req->host . $req->script . "?page=" . url($pagename) . "</guid>\n";
				$xml .= "<pubDate>" . html(date(DATE_RSS, $timestamp)) . "</pubDate>\n";
				$xml .= "</item>\n";
			}
			$xml .= "</channel>\n";
			$xml .= "</rss>";
			header("Content-type:application/rss+xml;charset=utf-8");
			die($xml);
		}
	}
	
	class Reference {
		private $reference = array();

		// store string and give back a unique reference
		private function put($string) {
			$hash = md5($string);
			$this->reference[$hash] = $string;
			return "{" . $hash . "}";
		}
			
		// take reference and give back the previousely stored string
		function get($m) {
			return $this->reference[$m[1]];
		}
		
		function pre ($m) {return $this->put('<pre>' . html($m[1]) . '</pre>');}
		function link ($m) {return $this->put('<a href="' . $m[2] . '" target="_blank">' . html($m[1]) . '</a>');}
		function img ($m) {return $this->put('<img alt="image" src="' . $m[1] . '" />');}
		function url ($m) {return $this->put('<a href="' . $m[1] . '" target="_blank">' . html($m[1]) . '</a>');}
		function hashtag ($m)  {return $this->put('<a href="index.php?action=hashtag&amp;query=' . url($m[1]) . '">' . html($m[1]) . '</a>');}
		function pagename ($m) {return $this->put('<a href="index.php?page=' . url($m[1]) . '">' . html($m[1]) . '</a>');}
	}
?>
